from marshy.types import ExternalItemType

from servey.servey_aws.event_handler.event_handler_abc import get_event_handlers, EventHandlerABC
from servey.servey_aws.router.router_abc import RouterABC


class AppsyncRouter(RouterABC):
    priority: int = 110

    def create_handler(self, event: ExternalItemType, context) -> EventHandlerABC:
        path = event.get('path', None)  # Diff appsync events
        if path is None:
            return
        action = self.find_action_for_path(path)
        handlers = get_event_handlers(action)
        for handler in handlers:
            if handler.is_usable(event, context):
                return handler



from enum import EnumMeta
from unittest import TestCase

from dateutil.relativedelta import relativedelta

from servey.servey_strawberry.entity_factory.dataclass_factory import _UserCache


class TestDataclassFactory(TestCase):
    def test_user_cache(self):
        cache = _UserCache(relativedelta(seconds=10))
        cache.set('foo', 'bar')
        self.assertEqual(cache.get('foo'), 'bar')
        cache.delete('foo')
        self.assertIsNone(cache.get('foo'))
        cache.set('foo', 'bar')
        cache.clear()
        self.assertIsNone(cache.get('foo'))



from abc import ABC, abstractmethod
from typing import Tuple, Optional

from marshy.factory.impl_marshaller_factory import get_impls
from marshy.types import ExternalItemType

from servey.action.action import Action
from servey.finder.action_finder_abc import find_actions_with_trigger_type
from servey.servey_aws.event_handler.event_handler_abc import EventHandlerABC
from servey.trigger.web_trigger import WebTrigger


class RouterABC(ABC):
    """ A router will attempt to create an event handler given a specific lambda event """
    priority: int = 100

    @abstractmethod
    def create_handler(self, event: ExternalItemType, context) -> EventHandlerABC:
        """ Create a handler for the event given """

    @property
    def web_trigger_actions(self) -> Tuple[Tuple[Action, WebTrigger], ...]:
        web_trigger_actions = getattr(self, '_web_trigger_actions', None)
        if web_trigger_actions is None:
            web_trigger_actions = list(find_actions_with_trigger_type(WebTrigger))
            web_trigger_actions.sort(key=lambda n: len(n[1].path or f"/actions/{n[0].name}"), reverse=True)
            web_trigger_actions = tuple(web_trigger_actions)
            setattr(self, '_web_trigger_actions', web_trigger_actions)
        return web_trigger_actions

    def find_action_for_path(self, path: str) -> Optional[Action]:
        for action, trigger in self.web_trigger_actions:
            action_path = trigger.path or f"/actions/{action.name}"
            if action_path == path:
                return action


def find_routers() -> Tuple[RouterABC]:
    routers = [router() for router in get_impls(RouterABC)]
    routers.sort(key=lambda r: r.priority, reverse=True)
    return tuple(routers)



from unittest import TestCase
from unittest.mock import patch

from servey.action.action import action, get_action
from servey.errors import ServeyError
from servey.servey_aws.lambda_router import invoke
from servey.trigger.web_trigger import WEB_GET


class TestLambdaInvoker(TestCase):
    def test_router(self):

        @action(triggers=WEB_GET)
        def echo_get(val: str) -> str:
            return val

        with (
            patch(
                "servey.servey_aws.router.router.find_actions",
                return_value=[get_action(echo_get)],
            )
        ):
            event = dict(action_name="echo_get", params=dict(val="foo"))
            result = invoke(event, None)
            expected_result = "foo"
            self.assertEqual(expected_result, result)

    def test_router_no_handler(self):

        @action(triggers=WEB_GET)
        def echo_get(val: str) -> str:
            """ Dummy """

        with (
            patch(
                "servey.servey_aws.router.router.find_actions",
                return_value=[get_action(echo_get)],
            ),
            self.assertRaises(StopIteration)
        ):
            event = dict(action_name="not_existing")
            invoke(event, None)

    def test_router_no_action(self):

        @action(triggers=WEB_GET)
        def echo_get(val: str) -> str:
            """ Dummy """

        with (
            patch(
                "servey.servey_aws.router.router.find_actions",
                return_value=[get_action(echo_get)],
            ),
            self.assertRaises(ServeyError)
        ):
            invoke({}, None)

    def test_router_api_gateway(self):

        @action(triggers=WEB_GET)
        def echo_get(val: str) -> str:
            return val

        with (
            patch(
                "servey.servey_aws.router.router.find_actions",
                return_value=[get_action(echo_get)],
            ),
            patch(
                "servey.finder.action_finder_abc.find_actions",
                return_value=[get_action(echo_get)],
            )
        ):
            event = dict(path="/actions/echo_get", params=dict(val="foo"))
            result = invoke(event, None)
            expected_result = "foo"
            self.assertEqual(expected_result, result)

    def test_router_appsync(self):

        @action(triggers=WEB_GET)
        def echo_get(val: str) -> str:
            return val

        with (
            patch(
                "servey.servey_aws.router.router.find_actions",
                return_value=[get_action(echo_get)],
            ),
            patch(
                "servey.finder.action_finder_abc.find_actions",
                return_value=[get_action(echo_get)],
            )
        ):
            event = dict(path="/actions/echo_get", params=dict(val="foo"))
            result = invoke(event, None)
            expected_result = "foo"
            self.assertEqual(expected_result, result)
